"use strict";
/**********************************************************************************
 * MIT License                                                                    *
 *                                                                                *
 * Copyright (c) 2021 Hyperjump Technology                                        *
 *                                                                                *
 * Permission is hereby granted, free of charge, to any person obtaining a copy   *
 * of this software and associated documentation files (the "Software"), to deal  *
 * in the Software without restriction, including without limitation the rights   *
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell      *
 * copies of the Software, and to permit persons to whom the Software is          *
 * furnished to do so, subject to the following conditions:                       *
 *                                                                                *
 * The above copyright notice and this permission notice shall be included in all *
 * copies or substantial portions of the Software.                                *
 *                                                                                *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR     *
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,       *
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE    *
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER         *
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,  *
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE  *
 * SOFTWARE.                                                                      *
 **********************************************************************************/
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const core_1 = require("@oclif/core");
const p_event_1 = __importDefault(require("p-event"));
const config_1 = require("../components/config");
const sort_1 = require("../components/config/sort");
const logger_1 = require("../components/logger");
const flush_1 = require("../components/logger/flush");
const history_1 = require("../components/logger/history");
const startup_message_1 = require("../components/logger/startup-message");
const schedule_notification_1 = require("../components/notification/schedule-notification");
const start_message_1 = require("../components/notification/start-message");
const context_1 = require("../context");
const events_1 = __importDefault(require("../events"));
const flag_1 = require("../flag");
const summary_notification_1 = require("../jobs/summary-notification");
const loaders_1 = __importDefault(require("../loaders"));
const looper_1 = require("../looper");
const symon_1 = __importDefault(require("../symon"));
const events_2 = require("../utils/events");
const pino_1 = require("../utils/pino");
const em = (0, events_2.getEventEmitter)();
let symonClient;
class Monika extends core_1.Command {
    async catch(error) {
        super.catch(error);
        if (symonClient) {
            await symonClient.sendStatus({ isOnline: false });
        }
        if (error instanceof core_1.Errors.ExitError) {
            return core_1.Errors.handle(error);
        }
        if (error.message.includes('EEXIT: 0')) {
            // this is normal exit, for example after running with --version,
            // not an error so just quit immediately
            // eslint-disable-next-line n/no-process-exit, unicorn/no-process-exit
            process.exit(0);
        }
        throw error;
    }
    deprecationHandler(config) {
        const showDeprecateMsg = {
            query: false,
        };
        const checkedConfig = {
            ...config,
            probes: config.probes?.map((probe) => ({
                ...probe,
                requests: probe.requests?.map((request) => ({
                    ...request,
                    alert: request.alerts?.map((alert) => {
                        if (alert.query) {
                            showDeprecateMsg.query = true;
                            return { ...alert, assertion: alert.query };
                        }
                        return alert;
                    }),
                })),
                alerts: probe.alerts?.map((alert) => {
                    if (alert.query) {
                        showDeprecateMsg.query = true;
                        return { ...alert, assertion: alert.query };
                    }
                    return alert;
                }),
            })),
        };
        if (showDeprecateMsg.query) {
            pino_1.log.warn('"alerts.query" is deprecated. Please use "alerts.assertion"');
        }
        return checkedConfig;
    }
    getProbes({ config, flags }) {
        const sortedProbes = (0, sort_1.sortProbes)(config.probes, flags.id);
        return sortedProbes.map((probe) => (0, looper_1.sanitizeProbe)((0, config_1.isSymonModeFrom)(flags), probe));
    }
    async run() {
        const monika = await this.parse(Monika);
        const _flags = monika.flags;
        (0, context_1.setContext)({ flags: _flags });
        try {
            if (_flags['create-config']) {
                await (0, config_1.createConfig)(_flags);
                return;
            }
            await (0, history_1.openLogfile)();
            if (_flags.logs) {
                await (0, logger_1.printAllLogs)();
                await (0, history_1.closeLog)();
                return;
            }
            if (_flags.flush) {
                await (0, flush_1.flush)(_flags.force);
                await (0, history_1.closeLog)();
                return;
            }
            if (_flags.summary) {
                (0, summary_notification_1.printSummary)(this.config);
                await (0, history_1.closeLog)();
                return;
            }
            await (0, loaders_1.default)(_flags, this.config);
            if ((0, config_1.isSymonModeFrom)(_flags)) {
                symonClient = new symon_1.default(_flags);
                await symonClient.initiate();
            }
            let isFirstRun = true;
            for (;;) {
                const controller = new AbortController();
                const { signal } = controller;
                const config = (0, config_1.getConfig)();
                const notifications = config.notifications || [];
                const probes = this.getProbes({ config, flags: _flags });
                // emit the sanitized probe
                em.emit(events_1.default.config.sanitized, probes);
                // save some data into files for later
                (0, summary_notification_1.savePidFile)(_flags.config, config);
                this.deprecationHandler(config);
                (0, startup_message_1.logStartupMessage)({
                    config,
                    flags: _flags,
                    isFirstRun,
                });
                (0, looper_1.startProbing)({
                    notifications,
                    probes,
                    signal,
                });
                if ((0, context_1.getContext)().isTest) {
                    break;
                }
                (0, start_message_1.sendMonikaStartMessage)(notifications).catch((error) => pino_1.log.error(error.message));
                // schedule status update notification
                (0, schedule_notification_1.scheduleSummaryNotification)({ config, flags: _flags });
                isFirstRun = false;
                // block the loop until receives config updated event
                // eslint-disable-next-line no-await-in-loop
                await (0, p_event_1.default)(em, events_1.default.config.updated);
                controller.abort('Monika configuration updated');
            }
        }
        catch (error) {
            await (0, history_1.closeLog)();
            this.error(error?.message, { exit: 1 });
        }
    }
}
Monika.description = 'Monika command line monitoring tool';
Monika.examples = [
    'monika',
    'monika --logs',
    'monika -r 1 --id "weather, stocks, 5, 7"',
    'monika --create-config',
    'monika --config https://raw.githubusercontent.com/hyperjumptech/monika/main/monika.example.yml --config-interval 900',
];
Monika.flags = {
    'auto-update': core_1.Flags.string({
        description: 'Enable auto-update for Monika. Available options: major, minor, patch. This will make Monika terminate itself on successful update but does not restart',
    }),
    config: core_1.Flags.string({
        char: 'c',
        default: flag_1.monikaFlagsDefaultValue.config,
        description: 'JSON configuration filename or URL. If none is supplied, will look for monika.yml in the current directory',
        env: 'MONIKA_JSON_CONFIG',
        multiple: true,
    }),
    'config-filename': core_1.Flags.string({
        default: flag_1.monikaFlagsDefaultValue['config-filename'],
        dependsOn: ['config'],
        description: 'The configuration filename for config file created if there is no config file found ',
    }),
    'config-interval': core_1.Flags.integer({
        default: flag_1.monikaFlagsDefaultValue['config-interval'],
        dependsOn: ['config'],
        description: 'The interval (in seconds) for periodic config checking if url is used as config source',
    }),
    'create-config': core_1.Flags.boolean({
        description: 'Create config from HAR (-H), postman (-p), insomnia (-I), sitemap (--sitemap), textfile (--text) export file, or open Monika Configuration Generator using default browser',
    }),
    'native-fetch': core_1.Flags.boolean({
        default: flag_1.monikaFlagsDefaultValue['native-fetch'],
        description: 'Use native fetch Node.js API instead of Axios for HTTP client',
    }),
    flush: core_1.Flags.boolean({
        description: 'Flush logs',
    }),
    'follow-redirects': core_1.Flags.integer({
        default: flag_1.monikaFlagsDefaultValue['follow-redirects'],
        description: 'Monika will follow redirects as many times as the specified value here. By default, Monika will follow redirects once. To disable redirects following, set the value to zero.',
    }),
    force: core_1.Flags.boolean({
        default: false,
        description: 'Force commands with a yes whenever Y/n is prompted.',
    }),
    har: core_1.Flags.string({
        char: 'H',
        description: 'Run Monika using a HAR file',
        exclusive: ['postman', 'insomnia', 'sitemap', 'text'],
        multiple: false,
    }),
    help: core_1.Flags.help({ char: 'h' }),
    id: core_1.Flags.string({
        char: 'i',
        description: 'Define specific probe ids to run',
        multiple: false,
    }),
    ignoreInvalidTLS: core_1.Flags.boolean({
        description: 'Configures whether HTTPS requests should ignore invalid certificates',
        default: false,
    }),
    insomnia: core_1.Flags.string({
        char: 'I',
        description: 'Run Monika using an Insomnia json/yaml file',
        exclusive: ['har', 'postman', 'sitemap', 'text'],
        multiple: false,
    }),
    'keep-verbose-logs': core_1.Flags.boolean({
        default: false,
        description: 'Store all requests logs to database',
    }),
    logs: core_1.Flags.boolean({
        char: 'l',
        description: 'Print all logs.',
    }),
    'one-probe': core_1.Flags.boolean({
        dependsOn: ['sitemap'],
        description: 'One Probe',
    }),
    output: core_1.Flags.string({
        char: 'o',
        description: 'Write monika config file to this file',
        multiple: false,
    }),
    postman: core_1.Flags.string({
        char: 'p',
        description: 'Run Monika using a Postman json file.',
        exclusive: ['har', 'insomnia', 'sitemap', 'text'],
        multiple: false,
    }),
    prometheus: core_1.Flags.integer({
        description: 'Specifies the port the Prometheus metric server is listening on. e.g., 3001. (EXPERIMENTAL)',
        exclusive: ['r'],
    }),
    repeat: core_1.Flags.integer({
        char: 'r',
        default: 0,
        description: 'Repeats the test run n times',
        multiple: false,
    }),
    retryInitialDelayMs: flag_1.retryInitialDelayMs,
    retryMaxDelayMs: flag_1.retryMaxDelayMs,
    sitemap: core_1.Flags.string({
        description: 'Run Monika using a Sitemap xml file.',
        exclusive: ['har', 'insomnia', 'postman', 'text'],
        multiple: false,
    }),
    'status-notification': core_1.Flags.string({
        description: 'Cron syntax for status notification schedule',
    }),
    stun: core_1.Flags.integer({
        char: 's',
        default: flag_1.monikaFlagsDefaultValue.stun,
        description: 'Interval in seconds to check STUN server',
        multiple: false,
    }),
    summary: core_1.Flags.boolean({
        default: false,
        description: 'Display a summary of monika running stats',
    }),
    'symon-api-version': (0, flag_1.symonAPIVersion)(),
    symonKey: core_1.Flags.string({
        dependsOn: ['symonUrl'],
        description: 'API Key for Symon',
    }),
    symonGetProbesIntervalMs: flag_1.symonGetProbesIntervalMs,
    symonLocationId: core_1.Flags.string({
        dependsOn: ['symonKey', 'symonUrl'],
        description: 'Location ID for Symon (optional)',
        required: false,
    }),
    symonMonikaId: core_1.Flags.string({
        dependsOn: ['symonKey', 'symonUrl'],
        description: 'Monika ID for Symon (optional)',
        required: false,
    }),
    symonReportInterval: core_1.Flags.integer({
        dependsOn: ['symonKey', 'symonUrl'],
        description: 'Interval for reporting to Symon in milliseconds (optional)',
        required: false,
    }),
    symonReportLimit: core_1.Flags.integer({
        dependsOn: ['symonKey', 'symonUrl'],
        description: 'Data limit to be reported to Symon (optional)',
        required: false,
    }),
    symonUrl: core_1.Flags.string({
        dependsOn: ['symonKey'],
        description: 'URL of Symon',
        hidden: false,
    }),
    text: core_1.Flags.string({
        description: 'Run Monika using a Simple text file',
        exclusive: ['postman', 'insomnia', 'sitemap', 'har'],
        multiple: false,
    }),
    verbose: core_1.Flags.boolean({
        default: false,
        description: 'Show verbose log messages',
    }),
    version: core_1.Flags.version({ char: 'v' }),
};
Monika.id = 'monika';
exports.default = Monika;
/**
 * Show Exit Message
 */
process.on('SIGINT', async () => {
    if (!process.env.DISABLE_EXIT_MESSAGE) {
        pino_1.log.info('Thank you for using Monika!');
        pino_1.log.info('We need your help to make Monika better.');
        pino_1.log.info('Can you give us some feedback by clicking this link https://github.com/hyperjumptech/monika/discussions?\n');
    }
    if (symonClient) {
        await symonClient.sendStatus({ isOnline: false });
        await symonClient.stop();
    }
    em.emit(events_1.default.application.terminated);
    // eslint-disable-next-line n/no-process-exit
    process.exit(process.exitCode);
});
//# sourceMappingURL=monika.js.map