"use strict";
/**********************************************************************************
 * MIT License                                                                    *
 *                                                                                *
 * Copyright (c) 2021 Hyperjump Technology                                        *
 *                                                                                *
 * Permission is hereby granted, free of charge, to any person obtaining a copy   *
 * of this software and associated documentation files (the "Software"), to deal  *
 * in the Software without restriction, including without limitation the rights   *
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell      *
 * copies of the Software, and to permit persons to whom the Software is          *
 * furnished to do so, subject to the following conditions:                       *
 *                                                                                *
 * The above copyright notice and this permission notice shall be included in all *
 * copies or substantial portions of the Software.                                *
 *                                                                                *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR     *
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,       *
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE    *
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER         *
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,  *
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE  *
 * SOFTWARE.                                                                      *
 **********************************************************************************/
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.isSymonModeFrom = exports.createConfig = exports.setupConfig = exports.updateConfig = exports.getConfig = void 0;
const chokidar_1 = require("chokidar");
const core_1 = require("@oclif/core");
const fs_1 = require("fs");
const is_url_1 = __importDefault(require("is-url"));
const events_1 = __importDefault(require("../../events"));
const context_1 = require("../../context");
const flag_1 = require("../../flag");
const events_2 = require("../../utils/events");
const hash_1 = require("../../utils/hash");
const open_website_1 = require("../../utils/open-website");
const pino_1 = require("../../utils/pino");
const parse_1 = require("./parse");
const validate_1 = require("./validate");
const create_config_1 = require("./create-config");
const js_yaml_1 = __importDefault(require("js-yaml"));
const process_1 = require("process");
const get_1 = require("./get");
const probe_1 = require("./probe");
const catch_error_handler_1 = require("../../utils/catch-error-handler");
const emitter = (0, events_2.getEventEmitter)();
const defaultConfigs = [];
let nonDefaultConfig;
const getConfig = () => {
    const { config, flags } = (0, context_1.getContext)();
    if (!config) {
        if (!isSymonModeFrom(flags)) {
            throw new Error('Configuration setup has not been run yet');
        }
        return {
            probes: [],
        };
    }
    return { ...config, probes: (0, probe_1.getProbes)() };
};
exports.getConfig = getConfig;
const updateConfig = async (config) => {
    pino_1.log.info('Updating config');
    try {
        const validatedConfig = await (0, validate_1.validateConfig)(config);
        const version = (0, hash_1.md5Hash)(validatedConfig);
        const hasChangeConfig = (0, context_1.getContext)().config?.version !== version;
        if (!hasChangeConfig) {
            return;
        }
        const newConfig = addConfigVersion(validatedConfig);
        (0, context_1.setContext)({ config: newConfig });
        (0, probe_1.setProbes)(newConfig.probes);
        emitter.emit(events_1.default.config.updated);
        pino_1.log.info('Config file update detected');
    }
    catch (error) {
        const message = (0, catch_error_handler_1.getErrorMessage)(error);
        if ((0, context_1.getContext)().isTest) {
            // return error during tests
            throw new Error(message);
        }
        pino_1.log.error(message);
        (0, process_1.exit)(1);
    }
};
exports.updateConfig = updateConfig;
const setupConfig = async (flags) => {
    const validFlag = await createConfigIfEmpty(flags);
    const config = await (0, get_1.getConfigFrom)(validFlag);
    await (0, exports.updateConfig)(config);
    watchConfigsChange(validFlag);
};
exports.setupConfig = setupConfig;
function addConfigVersion(config) {
    if (config.version) {
        return config;
    }
    const version = config.version || (0, hash_1.md5Hash)(config);
    return { ...config, version };
}
async function createConfigIfEmpty(flags) {
    // check for default config path when -c/--config not provided
    const hasConfig = flags.config.length > 0 ||
        flags.har ||
        flags.postman ||
        flags.insomnia ||
        flags.sitemap ||
        flags.text;
    if (!hasConfig) {
        pino_1.log.info('No Monika configuration available, initializing...');
        const configFilename = await (0, create_config_1.createConfigFile)(flags);
        return { ...flags, config: [configFilename] };
    }
    return flags;
}
async function watchConfigsChange(flags) {
    await Promise.all(flags.config.map((source, index) => watchConfigChange({
        flags,
        interval: flags['config-interval'] ||
            flag_1.monikaFlagsDefaultValue['config-interval'],
        source,
        type: 'monika',
        index,
    })));
}
function watchConfigChange({ flags, interval, source, type, index, }) {
    if ((0, is_url_1.default)(source)) {
        scheduleRemoteConfigFetcher({
            configType: type,
            interval,
            url: source,
            index,
        });
        return;
    }
    watchConfigFile({
        flags,
        path: source,
    });
}
function scheduleRemoteConfigFetcher({ configType, interval, url, index, }) {
    setInterval(async () => {
        try {
            const newConfig = await (0, parse_1.parseConfig)(url, configType);
            if (index === undefined) {
                nonDefaultConfig = newConfig;
            }
            else {
                defaultConfigs[index] = newConfig;
            }
            await (0, exports.updateConfig)((0, get_1.mergeConfigs)(defaultConfigs, nonDefaultConfig));
        }
        catch (error) {
            pino_1.log.error((0, catch_error_handler_1.getErrorMessage)(error));
        }
    }, interval * 1000);
}
function watchConfigFile({ flags, path }) {
    const isWatchConfigFile = !((0, context_1.getContext)().isTest || flags.repeat !== 0);
    if (isWatchConfigFile) {
        const watcher = (0, chokidar_1.watch)(path);
        watcher.on('change', async () => {
            const config = await (0, get_1.getConfigFrom)(flags);
            await (0, exports.updateConfig)(config);
        });
    }
}
const getPathAndTypeFromFlag = (flags) => {
    let path = flags.config?.[0];
    let type = 'monika';
    if (flags.postman) {
        path = flags.postman;
        type = 'postman';
    }
    if (flags.har) {
        path = flags.har;
        type = 'har';
    }
    if (flags.insomnia) {
        path = flags.insomnia;
        type = 'insomnia';
    }
    if (flags.sitemap) {
        path = flags.sitemap;
        type = 'sitemap';
    }
    if (flags.text) {
        path = flags.text;
        type = 'text';
    }
    return {
        path,
        type,
    };
};
const createConfig = async (flags) => {
    if (!flags.har &&
        !flags.postman &&
        !flags.insomnia &&
        !flags.sitemap &&
        !flags.text) {
        pino_1.log.info('Opening Monika Configuration Generator in your default browser...');
        (0, open_website_1.open)('https://hyperjumptech.github.io/monika-config-generator/');
    }
    else {
        const { path, type } = getPathAndTypeFromFlag(flags);
        if (!(0, fs_1.existsSync)(path)) {
            pino_1.log.error(`Couldn't found the ${path} file.`);
            return;
        }
        const parse = await (0, parse_1.parseConfig)(path, type, flags);
        const result = (0, get_1.addDefaultNotifications)(parse);
        const file = flags.output || 'monika.yml';
        if ((0, fs_1.existsSync)(file) && !flags.force) {
            const ans = await core_1.ux.ux.prompt(`\n${file} file is already exists. Overwrite (Y/n)?`);
            if (ans.toLowerCase() !== 'y') {
                pino_1.log.warn(`Command cancelled. You can use the -o flag to specify an output file or --force to overwrite without prompting.`);
                return;
            }
        }
        const yamlDoc = js_yaml_1.default.dump(result);
        (0, fs_1.writeFileSync)(file, yamlDoc, 'utf8');
        pino_1.log.info(`${file} file has been created.`);
    }
};
exports.createConfig = createConfig;
function isSymonModeFrom({ symonKey, symonUrl, }) {
    return Boolean(symonUrl) && Boolean(symonKey);
}
exports.isSymonModeFrom = isSymonModeFrom;
//# sourceMappingURL=index.js.map