"use strict";
/**********************************************************************************
 * MIT License                                                                    *
 *                                                                                *
 * Copyright (c) 2021 Hyperjump Technology                                        *
 *                                                                                *
 * Permission is hereby granted, free of charge, to any person obtaining a copy   *
 * of this software and associated documentation files (the "Software"), to deal  *
 * in the Software without restriction, including without limitation the rights   *
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell      *
 * copies of the Software, and to permit persons to whom the Software is          *
 * furnished to do so, subject to the following conditions:                       *
 *                                                                                *
 * The above copyright notice and this permission notice shall be included in all *
 * copies or substantial portions of the Software.                                *
 *                                                                                *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR     *
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,       *
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE    *
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER         *
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,  *
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE  *
 * SOFTWARE.                                                                      *
 **********************************************************************************/
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const test_1 = require("@oclif/test");
const _1 = require(".");
const context_1 = require("../../context");
const events_1 = __importDefault(require("../../events"));
const hash_1 = require("../../utils/hash");
const events_2 = require("../../utils/events");
const validation_1 = require("./validation");
const catch_error_handler_1 = require("../../utils/catch-error-handler");
describe('getConfig', () => {
    beforeEach(() => {
        (0, context_1.resetContext)();
    });
    it('should throw error when config is empty', () => {
        (0, test_1.expect)(() => (0, _1.getConfig)()).to.throw('Configuration setup has not been run yet');
    });
    it('should not throw error when config is empty in symon mode', () => {
        // arrange
        (0, context_1.setContext)({
            flags: {
                symonKey: 'bDF8j',
                symonUrl: 'https://example.com',
            },
        });
        // act & assert
        (0, test_1.expect)((0, _1.getConfig)()).deep.eq({ probes: [] });
    });
    it('should return empty array when config is empty in Symon mode', () => {
        // arrange
        (0, context_1.setContext)({
            flags: {
                symonKey: 'bDF8j',
                symonUrl: 'https://example.com',
            },
        });
        // act and assert
        (0, test_1.expect)((0, _1.getConfig)()).deep.eq({ probes: [] });
    });
});
describe('isSymonModeFrom', () => {
    it('should return true', () => {
        // act
        const isSymonMode = (0, _1.isSymonModeFrom)({
            symonKey: 'secret-key',
            symonUrl: 'https://example.com',
        });
        // assert
        (0, test_1.expect)(isSymonMode).eq(true);
    });
    it('should return false when symon key and url is empty', () => {
        // act
        const isSymonMode = (0, _1.isSymonModeFrom)({
            symonKey: '',
            symonUrl: '',
        });
        // assert
        (0, test_1.expect)(isSymonMode).eq(false);
    });
    it('should return false when symon key is empty', () => {
        // act
        const isSymonMode = (0, _1.isSymonModeFrom)({
            symonKey: '',
            symonUrl: 'https://example.com',
        });
        // assert
        (0, test_1.expect)(isSymonMode).eq(false);
    });
    it('should return false when symon url is empty', () => {
        // act
        const isSymonMode = (0, _1.isSymonModeFrom)({
            symonKey: 'secret-key',
            symonUrl: '',
        });
        // assert
        (0, test_1.expect)(isSymonMode).eq(false);
    });
});
describe('updateConfig', () => {
    beforeEach(() => {
        (0, context_1.resetContext)();
    });
    it('should throw error when config is invalid', async () => {
        // arrange
        let errorMessage = '';
        try {
            // act
            await (0, _1.updateConfig)({ probes: [] });
        }
        catch (error) {
            errorMessage = (0, catch_error_handler_1.getErrorMessage)(error);
        }
        // assert
        (0, test_1.expect)(errorMessage).contain('Probes object does not exists or has length lower than 1!');
    });
    it('should update config', async () => {
        // arrange
        const config = {
            probes: [
                {
                    id: '1',
                    name: '',
                    interval: 1000,
                    requests: [
                        {
                            url: 'https://example.com',
                            body: '',
                            followRedirects: 21,
                            timeout: 1000,
                        },
                    ],
                    alerts: [],
                },
            ],
        };
        const validatedProbes = {
            ...config,
            probes: await (0, validation_1.validateProbes)(config.probes),
        };
        let configFromEmitter = '';
        const eventEmitter = (0, events_2.getEventEmitter)();
        eventEmitter.once(events_1.default.config.updated, (data) => {
            configFromEmitter = data;
        });
        // act
        await (0, _1.updateConfig)(config);
        // assert
        (0, test_1.expect)((0, context_1.getContext)().config).to.deep.eq({
            ...validatedProbes,
            version: (0, hash_1.md5Hash)(validatedProbes),
        });
        (0, test_1.expect)(configFromEmitter).not.eq('');
    });
    it('should should not update config if there is no changes', async () => {
        // arrange
        const config = {
            probes: [
                {
                    id: '1',
                    name: '',
                    interval: 1000,
                    requests: [
                        {
                            url: 'https://example.com',
                            body: '',
                            followRedirects: 21,
                            timeout: 1000,
                        },
                    ],
                    alerts: [],
                },
            ],
        };
        let configFromEmitter = '';
        const eventEmitter = (0, events_2.getEventEmitter)();
        const eventListener = (data) => {
            configFromEmitter = data;
        };
        // act
        await (0, _1.updateConfig)(config);
        eventEmitter.once(events_1.default.config.updated, eventListener);
        await (0, _1.updateConfig)(config);
        // assert
        (0, test_1.expect)(configFromEmitter).eq('');
        eventEmitter.removeListener(events_1.default.config.updated, eventListener);
    });
});
//# sourceMappingURL=index.test.js.map