"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.printSummary = exports.savePidFile = exports.getSummaryAndSendNotif = void 0;
/**********************************************************************************
 * MIT License                                                                    *
 *                                                                                *
 * Copyright (c) 2021 Hyperjump Technology                                        *
 *                                                                                *
 * Permission is hereby granted, free of charge, to any person obtaining a copy   *
 * of this software and associated documentation files (the "Software"), to deal  *
 * in the Software without restriction, including without limitation the rights   *
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell      *
 * copies of the Software, and to permit persons to whom the Software is          *
 * furnished to do so, subject to the following conditions:                       *
 *                                                                                *
 * The above copyright notice and this permission notice shall be included in all *
 * copies or substantial portions of the Software.                                *
 *                                                                                *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR     *
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,       *
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE    *
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER         *
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,  *
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE  *
 * SOFTWARE.                                                                      *
 **********************************************************************************/
const os_1 = require("os");
const format_1 = __importDefault(require("date-fns/format"));
const config_1 = require("../components/config");
const history_1 = require("../components/logger/history");
const monika_notification_1 = require("@hyperjumptech/monika-notification");
const alert_message_1 = require("../components/notification/alert-message");
const context_1 = require("../context");
const ip_1 = __importDefault(require("../utils/ip"));
const pino_1 = require("../utils/pino");
const public_ip_1 = require("../utils/public-ip");
const response_time_log_1 = require("../components/logger/response-time-log");
const fs_1 = __importDefault(require("fs"));
const events_1 = __importDefault(require("../events"));
const events_2 = require("../utils/events");
const catch_error_handler_1 = require("../utils/catch-error-handler");
const read_file_1 = require("../utils/read-file");
const eventEmitter = (0, events_2.getEventEmitter)();
async function getSummaryAndSendNotif() {
    const config = (0, config_1.getConfig)();
    const { notifications } = config;
    if (!notifications)
        return;
    try {
        const { userAgent } = (0, context_1.getContext)();
        const privateIpAddress = (0, ip_1.default)();
        const [summary, osName, monikaInstance] = await Promise.all([
            (0, history_1.getSummary)(),
            (0, alert_message_1.getOSName)(),
            (0, alert_message_1.getMonikaInstance)(privateIpAddress),
        ]);
        const { numberOfIncidents, numberOfProbes, numberOfRecoveries, numberOfSentNotifications, } = summary;
        const responseTimelogLifeTimeInHour = (0, response_time_log_1.getLogLifeTimeInHour)();
        const tweetMessage = createTweetMessage({
            averageResponseTime: response_time_log_1.averageResponseTime,
            numberOfIncidents,
            numberOfProbes,
            numberOfRecoveries,
        });
        (0, monika_notification_1.sendNotifications)(notifications, {
            subject: `Monika Status`,
            body: `Status Update ${(0, format_1.default)(new Date(), 'yyyy-MM-dd HH:mm:ss XXX')}
Host: ${monikaInstance})
Number of probes: ${numberOfProbes}
Maximum response time: ${response_time_log_1.maxResponseTime} ms in the last ${responseTimelogLifeTimeInHour} hours
Minimum response time: ${response_time_log_1.minResponseTime} ms in the last ${responseTimelogLifeTimeInHour} hours
Average response time: ${response_time_log_1.averageResponseTime} ms in the last ${responseTimelogLifeTimeInHour} hours
Incidents: ${numberOfIncidents} in the last 24 hours
Recoveries: ${numberOfRecoveries} in the last 24 hours
Notifications: ${numberOfSentNotifications}
OS: ${osName}
Version: ${userAgent}

${tweetMessage}
`,
            summary: `There are ${numberOfIncidents} incidents and ${numberOfRecoveries} recoveries in the last 24 hours. - ${userAgent} - ${osName}`,
            meta: {
                type: 'status-update',
                time: (0, format_1.default)(new Date(), 'yyyy-MM-dd HH:mm:ss XXX'),
                hostname: (0, os_1.hostname)(),
                privateIpAddress,
                publicIpAddress: public_ip_1.publicIpAddress,
                monikaInstance,
                maxResponseTime: response_time_log_1.maxResponseTime,
                minResponseTime: response_time_log_1.minResponseTime,
                averageResponseTime: response_time_log_1.averageResponseTime,
                responseTimelogLifeTimeInHour,
                version: userAgent,
                ...summary,
            },
        }).catch((error) => pino_1.log.error(`Summary notification: ${error.message}`));
        if ((0, response_time_log_1.checkIs24HourHasPassed)()) {
            (0, response_time_log_1.resetlogs)();
        }
    }
    catch (error) {
        pino_1.log.error(`Summary notification: ${(0, catch_error_handler_1.getErrorMessage)(error)}`);
    }
}
exports.getSummaryAndSendNotif = getSummaryAndSendNotif;
/**
 * readsPidFile reads a local monika.pid file and returns the information in it
 * @returns {object} PidObject is returned
 */
async function readPidFile() {
    const fileContent = await (0, read_file_1.readFile)('monika.pid', 'utf8');
    const json = JSON.parse(fileContent);
    return {
        monikaPid: json.monikaPid,
        monikaConfigFile: json.monikaConfigFile,
        monikaStartTime: json.monikaStartTime,
        monikaProbes: json.monikaProbes,
        monikaNotifs: json.monikaNotifs,
    };
}
/**
 * savePidFile saves a monika.pid file with some useful information
 * @param {string} configFile is the configuration file used
 * @param {obj} config is a Config object
 * @returns void
 */
function savePidFile(configFile, config) {
    const data = JSON.stringify({
        monikaStartTime: new Date(),
        monikaConfigFile: configFile,
        monikaPid: process.pid,
        monikaProbes: config.probes ? config.probes.length : '0',
        monikaNotifs: config.notifications ? config.notifications.length : '0',
    });
    fs_1.default.writeFile('monika.pid', data, (err) => {
        if (err) {
            pino_1.log.info("couldn't save monika.pid file, got err: ", err);
        }
    });
}
exports.savePidFile = savePidFile;
// do somee cleanups on exit
eventEmitter.on(events_1.default.application.terminated, async () => {
    fs_1.default.unlink('monika.pid', (err) => {
        if (err) {
            pino_1.log.info('trying to cleanup monika.pid, got err: ', err);
        }
    });
});
/**
 * getDaysHours breaks down the date into days  hours minute string
 * @param {Date} startTime is the start time
 * @returns {string} text of the date to print
 */
function getDaysHours(startTime) {
    let duration = Math.abs((Date.now() - new Date(startTime).getTime()) / 1000);
    const numDays = Math.floor(duration / 86400);
    duration -= numDays * 86400; // get the remaining hours
    const numHours = Math.floor(duration / 3600) % 24;
    duration -= numHours * 3600; // get the remaining minutes
    const numMinutes = Math.floor(duration / 60) % 60;
    const numSeconds = Math.floor(duration - numMinutes * 60);
    return `${numDays} days, ${numHours} hours, ${numMinutes} minutes, ${numSeconds} seconds`;
}
/**
 * printSummary gathers and print some stats
 * @param {object} cliConfig is oclif config structure
 * @returns Promise<void>
 */
async function printSummary(cliConfig) {
    try {
        const pidObject = await readPidFile();
        const summary = await (0, history_1.getSummary)();
        const uptime = getDaysHours(pidObject.monikaStartTime);
        const host = `${(0, os_1.hostname)()} (${[public_ip_1.publicIpAddress, (0, ip_1.default)()]
            .filter(Boolean)
            .join('/')})`;
        pino_1.log.info(`Monika Summary \n
    Monika process id \t\t: ${pidObject.monikaPid}
    Active config file \t\t: ${pidObject.monikaConfigFile}
    Probes set \t\t\t: ${pidObject.monikaProbes}
    Notifications set \t\t: ${pidObject.monikaNotifs}
    Number of incidents \t: ${summary.numberOfIncidents} in last 24hr
    Number of recoveries \t: ${summary.numberOfRecoveries} in last 24hr
    Number of notifications \t: ${summary.numberOfSentNotifications} in last 24h

    Up time \t: ${uptime}
    Running on \t: ${host}
    App version : ${cliConfig.userAgent}

    `);
    }
    catch (error) {
        pino_1.log.error(`Summary notification: ${(0, catch_error_handler_1.getErrorMessage)(error)}`);
    }
}
exports.printSummary = printSummary;
function createTweetMessage({ averageResponseTime, numberOfIncidents, numberOfProbes, numberOfRecoveries, }) {
    const message = `I am using Monika by @hyperjump_tech to monitor ${numberOfProbes} probes! In the last 24 hours,

⏱ the average response time is ${averageResponseTime} ms
⚠️ there were ${numberOfIncidents} incidents
✅ and ${numberOfRecoveries} recoveries!

Give it a try!

https://monika.hyperjump.tech`;
    return `<a href=https://twitter.com/intent/tweet?text=${encodeURI(message)}&hashtags=opensource,monika>Tweet this status!</a>`;
}
//# sourceMappingURL=summary-notification.js.map