"use strict";
/**********************************************************************************
 * MIT License                                                                    *
 *                                                                                *
 * Copyright (c) 2021 Hyperjump Technology                                        *
 *                                                                                *
 * Permission is hereby granted, free of charge, to any person obtaining a copy   *
 * of this software and associated documentation files (the "Software"), to deal  *
 * in the Software without restriction, including without limitation the rights   *
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell      *
 * copies of the Software, and to permit persons to whom the Software is          *
 * furnished to do so, subject to the following conditions:                       *
 *                                                                                *
 * The above copyright notice and this permission notice shall be included in all *
 * copies or substantial portions of the Software.                                *
 *                                                                                *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR     *
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,       *
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE    *
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER         *
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,  *
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE  *
 * SOFTWARE.                                                                      *
 **********************************************************************************/
Object.defineProperty(exports, "__esModule", { value: true });
exports.PrometheusCollector = void 0;
const prom_client_1 = require("prom-client");
const request_1 = require("../../../interfaces/request");
let prometheusCustomCollector;
class PrometheusCollector {
    constructor() {
        // remove all registered metrics
        prom_client_1.register.clear();
        // register metric collector
        const statusCode = new prom_client_1.Gauge({
            name: 'monika_request_status_code_info',
            help: 'HTTP status code',
            labelNames: ['id', 'name', 'url', 'method'],
        });
        const probeResult = new prom_client_1.Gauge({
            name: 'monika_probe_result',
            help: 'Probe result: -1=unknown, 0=failed, 1=success',
            labelNames: ['id', 'name', 'url', 'method'],
        });
        const responseTime = new prom_client_1.Histogram({
            name: 'monika_request_response_time_seconds',
            help: 'Duration of probe request in seconds',
            labelNames: [
                'id',
                'name',
                'url',
                'method',
                'statusCode',
                'result',
            ],
        });
        const responseSize = new prom_client_1.Gauge({
            name: 'monika_request_response_size_bytes',
            help: 'Size of response size in bytes',
            labelNames: [
                'id',
                'name',
                'url',
                'method',
                'statusCode',
                'result',
            ],
        });
        const alertTriggeredTotal = new prom_client_1.Counter({
            name: 'monika_alert_total',
            help: 'Total alert triggered',
            labelNames: ['id', 'name', 'url', 'method', 'alertQuery'],
        });
        const probeRunningTotal = new prom_client_1.Gauge({
            name: 'monika_probe_running_total',
            help: 'Total of probe running',
            labelNames: ['id'],
        });
        const probesTotal = new prom_client_1.Gauge({
            name: 'monika_probes_total',
            help: 'Total of all probe',
        });
        // register and collect default Node.js metrics
        (0, prom_client_1.collectDefaultMetrics)({ register: prom_client_1.register });
        prometheusCustomCollector = {
            statusCode,
            probeResult,
            responseTime,
            responseSize,
            alertTriggeredTotal,
            probeRunningTotal,
            probesTotal,
        };
    }
    collectProbeTotal(total) {
        if (!prometheusCustomCollector) {
            throw new Error('Prometheus collector is not registered');
        }
        prometheusCustomCollector.probesTotal.set(total);
    }
    decrementProbeRunningTotal(id) {
        if (!prometheusCustomCollector) {
            throw new Error('Prometheus collector is not registered');
        }
        prometheusCustomCollector.probeRunningTotal.labels(id).dec();
    }
    incrementProbeRunningTotal(id) {
        if (!prometheusCustomCollector) {
            throw new Error('Prometheus collector is not registered');
        }
        prometheusCustomCollector.probeRunningTotal.labels(id).inc();
    }
    resetProbeRunningTotal() {
        if (!prometheusCustomCollector) {
            throw new Error('Prometheus collector is not registered');
        }
        prometheusCustomCollector.probeRunningTotal.reset();
    }
    collectProbeRequestMetrics(probeResult) {
        if (!prometheusCustomCollector) {
            throw new Error('Prometheus collector is not registered');
        }
        const { probe, requestIndex, response } = probeResult;
        const { id, name, requests } = probe;
        if (!requests || requests.length === 0) {
            return;
        }
        const request = requests[requestIndex];
        const { method, url } = request;
        const { headers, responseTime, status } = response;
        const result = response.result ?? request_1.probeRequestResult.unknown;
        const milliSecond = 1000;
        const responseTimeInSecond = responseTime / milliSecond || 0;
        const responseSizeBytes = typeof headers === 'string'
            ? undefined
            : Number(headers['content-length']);
        const labels = {
            id,
            name,
            url,
            result,
            method: method ?? 'GET',
            statusCode: status,
        };
        const { statusCode, probeResult: probeResultCollector, responseTime: resposeTimeCollector, responseSize, } = prometheusCustomCollector;
        // collect metrics
        statusCode
            ?.labels({
            id,
            name,
            url,
            method: method ?? 'GET',
        })
            .set(status);
        probeResultCollector
            ?.labels({
            id,
            name,
            url,
            method: method ?? 'GET',
        })
            .set(result);
        resposeTimeCollector?.labels(labels).observe(responseTimeInSecond);
        responseSize?.labels(labels).set(responseSizeBytes || 0);
    }
    collectTriggeredAlert(probeResult) {
        if (!prometheusCustomCollector) {
            throw new Error('Prometheus collector is not registered');
        }
        const { alertQuery, probe, requestIndex } = probeResult;
        const { id, name, requests } = probe;
        if (!requests || requests.length === 0) {
            return;
        }
        const request = requests[requestIndex];
        const { method, url } = request;
        const labels = {
            id,
            name,
            url,
            method: method ?? 'GET',
            alertQuery,
        };
        const { alertTriggeredTotal } = prometheusCustomCollector;
        // collect metrics
        alertTriggeredTotal?.labels(labels).inc();
    }
}
exports.PrometheusCollector = PrometheusCollector;
//# sourceMappingURL=collector.js.map